<?php /* -*- PHP -*- */

/** Constant definitions **/

/* log-level:
   0 - no log message
   1 - error message
   2 - warning message
   3 - debug message
*/
define ("_LOG_NOLOG_", 0);
define ("_LOG_ERROR_", 1);
define ("_LOG_WARN_", 2);
define ("_LOG_DEBUG_", 3);

if (!defined ("_UNIT_TEST_"))
  require_once ("config.inc.php");

/** Function definitions **/

/* log function */

function logging ($level, $msg) {

  $msg .= "\n";

  if ($level <= LOGLEVEL) {
    if (strlen (LOGFILE) != 0)
      $fd = fopen (LOGFILE, "a");
    else
      $fd = STDERR;
    if ($fd) {
      $formatted_msg = date (DATE_RFC822) . ": " . $msg;
      $len = strlen ($formatted_msg);
      $nb = fwrite ($fd, $formatted_msg, $len);
      if (strlen (LOGFILE) != 0)
        fclose ($fd);
    }
  }

  return $msg;
}

/* analyse resource */

function analyse_resource ($resource) {

  $element = array ("contact" => null,
                    "description" => null,
                    "allow_read" => null,
                    "deny_read" => null,
                    "allow_push" => null,
                    "deny_push" => null,
                    "archive_bz2" => null,
                    "archive_gz" => null,
                    "archive_zip" => null,
                    "encoding" => null,
                    "hidden" => null,
                    "style" => null,
                    "accesslog" => null,
                    "errorlog" => null,
                    "rest" => "");

  $is_web_section_open = false;
  foreach (preg_split ('/\n/', $resource) as $line) {

    /* clean line */
    $line = trim ($line);
    logging (_LOG_DEBUG_, "line = '$line'");

    /* check section name */
    if (strcmp ($line, "[web]") == 0) {
      $is_web_section_open = true;
      logging (_LOG_DEBUG_, "open web section");
      continue;
    }
    if (strpos ($line, "[") === 0) {
      $is_web_section_open = false;
      logging (_LOG_DEBUG_, "close web section");
    }
    if (!$is_web_section_open) {
      if (strlen ($line) > 0)
        $element["rest"] .= "$line\n";
      continue;
    }

    /* look for scheme 'variable = value' */
    $pos = strpos ($line, "=");
    if ($pos === false) {
      if (strlen ($line) > 0) {
        logging (_LOG_WARN_, "unknown line '$line'");
        continue;
      }
    }
    $var = trim (substr ($line, 0, $pos - 1));
    $val = trim (substr ($line, $pos + 1));

    /* check variable */
    switch ($var) {

    case "contact":
      $element["contact"] = $val;
      break;

    case "description":
      $element["description"] = $val;
      break;

    case "allow_read":
      $element["allow_read"] = $val;
      break;

    case "deny_read":
      $element["deny_read"] = $val;
      break;

    case "allow_push":
      $element["allow_push"] = $val;
      break;

    case "deny_push":
      $element["deny_push"] = $val;
      break;

    case "allow_archive":
      foreach (preg_split('/[\s,]+/', $val) as $type) {
        logging (_LOG_DEBUG_, "type: '$type'");
        switch (trim ($type)) {
        case "bz2":
          $element["archive_bz2"] = 1;
          break;
        case "gz":
          $element["archive_gz"] = 1;
          break;
        case "zip":
          $element["archive_zip"] = 1;
          break;
        default:
          logging (_LOG_WARN_, "unknown archive type '$type'");
        }
      }
      break;

    case "encoding":
      $element["encoding"] = $val;
      break;

    case "hidden":
      $element["hidden"] = (strcmp ($val, "True") == 0) ? 1 : 0;
      break;

    case "style":
      $element["style"] = $val;
      break;

    case "accesslog":
      $element["accesslog"] = $val;
      break;

    case "errorlog":
      $element["errorlog"] = $val;
      break;

    default:
      if (strlen ($line) > 0) {
        logging (_LOG_WARN_, "unknown line '$line'");
        $element["rest"] .= "[web]\n$line\n";
      }
    }
  }

  if (strlen ($element["rest"]) == 0)
    unset ($element["rest"]);

  return $element;
}

/* format resource */

function format_resource ($element) {

  $resource = "";

  if (isset ($element["contact"]))
    $resource .= "contact = $element[contact]\n";

  if (isset ($element["description"]))
    $resource .= "description = $element[description]\n";

  if (isset ($element["allow_read"]))
    $resource .= "allow_read = $element[allow_read]\n";

  if (isset ($element["deny_read"]))
    $resource .= "deny_read = $element[deny_read]\n";

  if (isset ($element["allow_push"]))
    $resource .= "allow_push = $element[allow_push]\n";

  if (isset ($element["deny_push"]))
    $resource .= "deny_push = $element[deny_push]\n";

  $archive_type = "";
  if ((isset ($element["archive_bz2"])) && ($element["archive_bz2"]))
    $archive_type .= " bz2";
  if ((isset ($element["archive_gz"])) && ($element["archive_gz"]))
    $archive_type .= " gz";
  if ((isset ($element["archive_zip"])) && ($element["archive_zip"]))
    $archive_type .= " zip";
  if (strlen ($archive_type) > 0)
    $resource .= "allow_archive =$archive_type\n";

  if (isset ($element["encoding"]))
    $resource .= "encoding = $element[encoding]\n";

  if (isset ($element["hidden"]))
    $resource .= "hidden = " . ($element["hidden"] ? "True" : "False") . "\n";

  if (isset ($element["style"]))
    $resource .= "style = $element[style]\n";

  if (isset ($element["accesslog"]))
    $resource .= "accesslog = $element[accesslog]\n";

  if (isset ($element["errorlog"]))
    $resource .= "errorlog = $element[errorlog]\n";

  if (strlen ($resource) > 0)
    $resource = "[web]\n$resource";

  if (isset ($element["rest"]))
    $resource .= $element["rest"];

  return $resource;
}

/* build mercurial tree */

function build_hg_tree ($dir = HGREPOS) {

  $tab = array ();
  if (!is_dir ($dir)) return;

  $basename = preg_replace (',^' . HGREPOS . '[/]*,', "", $dir);
  if (strcmp ($basename, $dir) == 0) return;

  $dh = opendir ($dir);
  logging (_LOG_DEBUG_, "$dh = opendir ('$dir')");
  if (!$dh) return;

  $is_fisrt = true;
  while (($f = readdir ($dh)) != false) {
    logging (_LOG_DEBUG_, "'$f' = readdir ('$dh')");
    if ((strcmp ($f, ".") == 0) || (strcmp ($f, "..") == 0)) continue;

    $element = array ();

    $name = preg_replace (",^[/]*,", "", "$basename/$f");

    if (is_dir ("$dir/$f/.hg")) {
      $element["name"] = $name;

      $element["base"] = "_" . preg_replace ("/[^a-zA-Z0-9]/", "_", $name);

      $resource = "";
      $fname = "$dir/$f/.hg/hgrc";
      if (is_file ($fname)) {
        $fd = fopen ($fname, "r");
        logging (_LOG_DEBUG_, "$fd = fopen ('$fname', 'r')");
        if ($fd) {
          if ($len = filesize ($fname))
            $resource = fread ($fd, $len);
          logging (_LOG_DEBUG_, "'$resource' = fread ($fd, $len)");
          fclose ($fd);
          logging (_LOG_DEBUG_, "fclose ($fd)");
        }
      }
      $element = array_merge ($element, analyse_resource ($resource));
    }

    if (count ($element) > 0)
      array_push ($tab, $element);

    $child = build_hg_tree ("$dir/$f");
    if (count ($child))
      $tab = array_merge ($tab, $child);

  }

  closedir ($dh);

  return $tab;
}

/* create xml tree */

function create_xml_tree ($tree) {
  $buffer = "<xml>\n";

  if ($tree == null)
    $tree = array ();

  foreach ($tree as $element) {
    $buffer .= "<tree>\n";
    $buffer .= "<base>$element[base]</base>";
    $buffer .= "<name>$element[name]</name>";
    $buffer .= "<contact><![CDATA[$element[contact]]]></contact>";
    $buffer .= "<description><![CDATA[$element[description]]]></description>";
    $buffer .= "<allow_read><![CDATA[$element[allow_read]]]></allow_read>";
    $buffer .= "<deny_read><![CDATA[$element[deny_read]]]></deny_read>";
    $buffer .= "<allow_push><![CDATA[$element[allow_push]]]></allow_push>";
    $buffer .= "<deny_push><![CDATA[$element[deny_push]]]></deny_push>";
    if (isset ($element["archive_bz2"]))
    $buffer .= "<archive_bz2>$element[archive_bz2]</archive_bz2>";
    if (isset ($element["archive_gz"]))
    $buffer .= "<archive_gz>$element[archive_gz]</archive_gz>";
    if (isset ($element["archive_zip"]))
    $buffer .= "<archive_zip>$element[archive_zip]</archive_zip>";
    if (isset ($element["encoding"]))
    $buffer .= "<encoding>$element[encoding]</encoding>";
    if (isset ($element["hidden"]))
    $buffer .= "<hidden>$element[hidden]</hidden>";
    if (isset ($element["style"]))
      $buffer .= "<style>$element[style]</style>";
    if (isset ($element["accesslog"]))
      $buffer .= "<accesslog>$element[accesslog]</accesslog>";
    if (isset ($element["errorlog"]))
      $buffer .= "<errorlog>$element[errorlog]</errorlog>";
    $buffer .= "</tree>\n";
  }

  $buffer .= "</xml>\n";

  return $buffer;
}

/* create xml config */

function create_xml_config () {

  $name = $style = $auth = $authfile = $authldaphost = $authldaprequest = "";

  /* read access file */
  $access = '';
  $fname = HGREPOS . "/.htaccess";
  if (file_exists ($fname)) {
    $fd = fopen ($fname, "r");
    logging (_LOG_DEBUG_, "$fd = fopen ('$fname', 'r')");
    if ($fd) {
      if ($len = filesize ($fname))
        $access = fread ($fd, $len);
      logging (_LOG_DEBUG_, "'$access' = fread ($fd, $len)");
      fclose ($fd);
      logging (_LOG_DEBUG_, "fclose ($fd)");
    }
  }

  /* analyse access file */
  foreach (preg_split ('/\n/', $access) as $line) {

    /* clean line */
    $line = trim ($line);
    logging (_LOG_DEBUG_, "line = '$line'");

    /* look for url base name */
    if (preg_match ('/RewriteRule \/(.*)\/\(\.\*\) .*\/hgweb\.cgi\/\$1/', $line, $matches))
      $name = $matches[1];

    /* look for authentification provider */
    if (preg_match ("/AuthBasicProvider (.*)/", $line, $matches))
      $auth = $matches[1];

    /* look for authentification file */
    if (preg_match ("/AuthUserFile (.*)/", $line, $matches))
      $authfile = $matches[1];

    /* look for ldap parameters */
    if (preg_match ("/AuthLDAPURL \"ldap:\/\/(.*)\/(.*)\?.*\?.*/", $line, $matches)) {
      $authldaphost = $matches[1];
      $authldaprequest = $matches[2];
    }

  }

  /* read resource file */
  $resources = '';
  $fname = HGREPOS . "/hgweb.config";
  if (file_exists ($fname)) {
    $fd = fopen ($fname, "r");
    logging (_LOG_DEBUG_, "$fd = fopen ('$fname', 'r')");
    if ($fd) {
      if ($len = filesize ($fname))
        $resources = fread ($fd, $len);
      logging (_LOG_DEBUG_, "'$access' = fread ($fd, $len)");
      fclose ($fd);
      logging (_LOG_DEBUG_, "fclose ($fd)");
    }
  }

  /* analyse resource file */
  $res = analyse_resource ($resources);
  if (isset ($res["style"]))
    $style = $res["style"];

  /* output xml */
  $buffer = "<xml><config>\n";
  $buffer .= "<root><![CDATA[" . HGREPOS . "]]></root>\n";
  $buffer .= "<name><![CDATA[$name]]></name>\n";
  $buffer .= "<logfile><![CDATA[" . LOGFILE . "]]></logfile>\n";
  $buffer .= "<loglevel>" . LOGLEVEL . "</loglevel>\n";
  $buffer .= "<style>$style</style>\n";
  $buffer .= "<auth>$auth</auth>\n";
  $buffer .= "<authfile><![CDATA[$authfile]]></authfile>\n";
  $buffer .= "<authldaphost><![CDATA[$authldaphost]]></authldaphost>\n";
  $buffer .= "<authldaprequest><![CDATA[$authldaprequest]]></authldaprequest>\n";
  $buffer .= "</config></xml>\n";

  return $buffer;
}

/* search repos in tree */

function search_repos ($tree, $base) {
  foreach ($tree as $element)
    if (array_key_exists ("base", $element) &&
        is_string ($element["base"]) &&
        (strcmp ($base, $element["base"]) == 0))
      return $element;
  return null;
}

/* remove directory */

function unlink_dir ($name) {

  if (!file_exists ($name))
    return false;

  if (!is_dir ($name)) {
    $ret = unlink ($name);
    logging (_LOG_DEBUG_, "$ret = unlink ('$name')");
    return $ret;
  }

  $dh = opendir ($name);
  logging (_LOG_DEBUG_, "$dh = opendir ('$name')");
  if (!$dh) return false;

  $ret = true;
  while (($f = readdir ($dh)) != false) {
    logging (_LOG_DEBUG_, "'$f' = readdir ($dh)");
    if ((strcmp ($f, ".") == 0) || (strcmp ($f, "..") == 0)) continue;

    $subname = "$name/$f";
    if (is_dir ($subname))
      $ret &= unlink_dir ($subname);
    else {
      $ret_partial = @unlink ($subname);
      logging (_LOG_DEBUG_, "$ret_partial = unlink ('$subname')");
      $ret &= $ret_partial;
    }
  }
  $ret_partial = @rmdir ($name);
  logging (_LOG_DEBUG_, "$ret_partial = rmdir ('$name')");
  $ret &= $ret_partial;

  return $ret;
}

function file_write ($name, $string) {

  $status = -1;

  $fd = @fopen ($name, "w");
  logging (_LOG_DEBUG_, "$fd = fopen ('$name', 'w')");
  if ($fd) {
    $nb = fwrite ($fd, $string, $len = strlen ($string));
    logging (_LOG_DEBUG_, "$nb = fwrite ($fd, '$string', $len)");
    $status = ($len == $nb) ? 0 : 1;
    fclose ($fd);
    logging (_LOG_DEBUG_, "fclose ($fd)");
  }

  return $status;
}

function real_name ($name) {

  $path = array ();

  foreach (preg_split (",/,", $name) as $f) {
    if ((strlen ($f) == 0) || (strcmp ($f, ".") == 0))
      continue;
    if (strcmp ($f, "..") == 0) {
      if (array_pop ($path) == null) {
        return "";
      }
    } else {
      array_push ($path, $f);
    }
  }

  $name = array_reduce ($path, function ($x, $y) { return is_null ($x) ? $y : sprintf ("%s/%s", $x, $y); });

  return $name;
}

/* update configuration */

function update_configuration ($resources, &$msg) {

  $stat = 0;
  $updated = false;

  /* check resources */

  $root = HGREPOS;
  if (isset ($resources["root"]))
    $root = $resources["root"];

  $logfile = LOGFILE;
  if (isset ($resources["logfile"]))
    $logfile = $resources["logfile"];

  $loglevel = LOGLEVEL;
  if (isset ($resources["loglevel"]))
    $loglevel = $resources["loglevel"];

  $name = "hg";
  if (isset ($resources["name"]))
    $name = $resources["name"];
  $name = preg_replace (",^/*,", "", $name);

  $style = "monoblue";
  if (isset ($resources["style"]))
    $style = $resources["style"];

  $auth = "file";
  if (isset ($resources["auth"]))
    $auth = $resources["auth"];

  $authfile = "password.md5";
  if (isset ($resources["authfile"]))
    $authfile = $resources["authfile"];

  $authldaphost = "ldap.my_company.my_ext";
  if (isset ($resources["authldaphost"]))
    $authldaphost = $resources["authldaphost"];

  $authldaprequest = "ou=People,dc=my_company,dc=my_ext";
  if (isset ($resources["authldaprequest"]))
    $authldaprequest = $resources["authldaprequest"];

  /* check modifications */
  $config = "<?php /* -*- PHP -*- */\n/* Mercurial repository root */\ndefine ('HGREPOS', '$root');\n/* Log file name */\ndefine ('LOGFILE', '$logfile');\n/* Log level */\ndefine ('LOGLEVEL', $loglevel);\n?>";
  $fname = "config.inc.php";
  if ((!file_exists ($fname)) || (strcmp (file_get_contents ($fname), $config) != 0)) {
    switch (file_write ($fname, $config)) {
    case -1:
      $msg .= logging (_LOG_ERROR_, "Can't write configuration file");
      $stat = 2;
      break;
    case 1:
      $msg .= logging (_LOG_WARN_, "Configuration file not completely written");
      $stat = 1;
      break;
    default:
      $msg .= logging (_LOG_DEBUG_, "Configuration updated");
    }

    $updated = true;
  }

  /* create root repository */

  if (!file_exists ($root)) {

    $ret = @mkdir ($root, 0750, true);
    logging (_LOG_DEBUG_, "$ret = mkdir ('$root', 0750, true)");
    if (!$ret) {
      $msg .= logging (_LOG_ERROR_, "Can't create root repository");
      return 2;
    } else {
      $msg .= logging (_LOG_WARN_, "Create root repository '$root'");
    }

    $updated = true;
  }

  /* test repository root */

  if ((!file_exists ($root)) || (!is_dir ($root))) {

    $msg .= logging (_LOG_ERROR_, "Can't open root repository");
    $stat = 2;

  } else {

    /* allow cgi */
    $access = "Order allow,deny\nAllow from all\n";
    $access .= "\n";
    $access .= "Options ExecCGI FollowSymLinks\nAddHandler cgi-script .cgi\n";
    $access .= "\n";

    /* redirection */
    $access .= "RewriteEngine On\nRewriteBase /$name\nRewriteRule ^\$ hgweb.cgi [L]\nRewriteCond %{REQUEST_FILENAME} !-f\nRewriteRule (.*) hgweb.cgi/\$1 [QSA,L]\n";
    $access .= "\n";

    /* authentification */
    $access .= "AuthType Basic\nAuthName \"Mercurial repositories\"\n";
    switch ($auth) {
    case 'file':
      $access .= "AuthBasicProvider file\nAuthUserFile $authfile\nRequire valid-user\n";
      break;

    case 'ldap':
      $access .= "AuthBasicProvider ldap\nAuthzLDAPAuthoritative on\nAuthLDAPURL \"ldap://$authldaphost/$authldaprequest?uid?sub\"\nAuthLDAPRemoteUserIsDN off\nRequire ldap-filter &(uid=*)\n";
      break;
    }

    /* write htaccess */
    $fname = "$root/.htaccess";
    if ((!file_exists ($fname)) || (strcmp (file_get_contents ($fname), $access) != 0)) {
      switch (file_write ($fname, $access)) {
      case -1:
        $msg .= logging (_LOG_ERROR_, "Can't write access file");
        return 2;
        break;
      case 1:
        $msg .= logging (_LOG_WARN_, "Access file not completely written");
      $stat = 1;
      break;
      default:
        $msg .= logging (_LOG_DEBUG_, "Access file written");
      }

      $updated = true;
    }

    /* hgweb server */
    $hgweb_cgi = "#!/usr/bin/env python\nconfig = 'hgweb.config'\nfrom mercurial import demandimport; demandimport.enable()\nfrom mercurial.hgweb import hgweb, wsgicgi\napplication = hgweb(config)\nwsgicgi.launch(application)\n";
    $fname = "$root/hgweb.cgi";
    if ((!file_exists ($fname)) || (strcmp (file_get_contents ($fname), $hgweb_cgi) != 0)) {
      switch (file_write ($fname, $hgweb_cgi)) {
      case -1:
        $msg .= logging (_LOG_ERROR_, "Can't mercurial web server exec");
        return 2;
        break;
      case 1:
        $msg .= logging (_LOG_WARN_, "Mercurial web server exec not completely written");
        $stat = 1;
        break;
      default:
        $msg .= logging (_LOG_DEBUG_, "Mercurial web server exec written");
      }
      $ret = @chmod ("$root/hgweb.cgi", 0755);
      if (!$ret) {
        $msg .= logging (_LOG_WARN_, "Can't change Mercurial web server config mode");
        $stat = 1;
      }

      $updated = true;
    }

    /* hgweb config */
    $hgweb_config = "[collections]\n$root = $root\n";
    $hgweb_config .= "\n";
    $hgweb_config .= "[web]\nstyle = $style\npush_ssl = false\nbaseurl = /$name\n";
    $fname = "$root/hgweb.config";
    if ((!file_exists ($fname)) || (strcmp (file_get_contents ($fname), $hgweb_config) != 0)) {
      switch (file_write ($fname, $hgweb_config)) {
      case -1:
        $msg .= logging (_LOG_ERROR_, "Can't mercurial web server config");
        return 2;
        break;
      case 1:
        $msg .= logging (_LOG_WARN_, "Mercurial web server config not completely written");
        $stat = 1;
        break;
      default:
        $msg .= logging (_LOG_DEBUG_, "Mercurial web server config written");
      }

      $updated = true;
    }
  }

  if (!$updated) {
    $msg .= logging (_LOG_WARN_, "Nothing to update for configuration");
    $stat = 1;
  }

  return $stat;
}

/* create repository */

function create_repository ($hgtree, $name, $resources, &$msg) {

  $stat = 0;
  $err = -1;

  /* check parameters */
  $path = real_name ($name);
  if (strlen ($path) == 0) {
    $msg .= logging (_LOG_ERROR_, "Path '$name' is empty or outside Mercurail global repository");
    return 2;
  }
  $name = $path;
  logging (_LOG_DEBUG_, "name: '$name'");

  /* check if repository exists */
  if (file_exists (HGREPOS . "/$name/.hg")) {
    $msg .= logging (_LOG_ERROR_, "Mercurial repository '$name' already exists");
    return 2;
  }

  /* create mercurial repository */
  passthru ("hg init " . HGREPOS . "/$name", $err);
  logging (_LOG_DEBUG_, "passthru ('hg init " . HGREPOS . "/$name', $err)");
  if ($err != 0) {
    $msg .= logging (_LOG_ERROR_, "Can't create mercurial repository '$name'");
    return 2;
  }
  $msg .= logging (_LOG_WARN_, "Create repository '$name'");

  /* resource file */
  switch (file_write (HGREPOS . "/$name/.hg/hgrc", format_resource ($resources))) {
  case -1:
    $msg .= logging (_LOG_WARN_, "Can't write resource file for '$name'");
    $stat = 1;
    break;
  case 1:
    $msg .= logging (_LOG_WARN_, "Resource file for '$name' not completely written");
    $stat = 1;
    break;
  }

  return $stat;
}

/* delete repository */

function delete_repository ($hgtree, $base, &$msg) {

  $stat = 0;
  $err = -1;

  /* check if directory is a repository */
  $repos = search_repos ($hgtree, $base);
  $name = $repos["name"];
  if (file_exists (HGREPOS . "/$name/.hg") == true) {
    passthru ("hg -R " . HGREPOS . "/$name verify >/dev/null", $err);
    logging (_LOG_DEBUG_, "passthru ('hg -R " . HGREPOS . "/$name verify', $err)");
  }
  if ($err != 0) {
    $msg .= logging (_LOG_ERROR_, "Directory '$name' is not a mercurial repository");
    return 2;
  }

  /* remove repository */
  if (unlink_dir (HGREPOS . "/$name/.hg") == false) {
    $msg .= logging (_LOG_ERROR_, "Can't remove completely repository '$name'");
    return 2;
  }
  $msg .= logging (_LOG_WARN_, "Repository '$name' deleted");

  /* close directory if empty */
  $ret = @rmdir (HGREPOS . "/$name");
  logging (_LOG_DEBUG_, "$ret = rmdir ('" . HGREPOS . "/$name')");
  if (!$ret) {
    $msg .= logging (_LOG_WARN_, "Can't remove directory '$name' (not empty or protected)");
    $stat = 1;
  }

  return $stat;
}

/* edit repository */

function edit_repository ($hgtree, $base, $resources, &$msg) {

  $stat = 0;

  /* check if directory is a repository */
  $repos = search_repos ($hgtree, $base);
  if ($repos == null) {
    $msg .= logging (_LOG_ERROR_, "Can't find repository '$base'");
    return 2;
  }
  logging (_LOG_DEBUG_, "repos[name] = '$repos[name]'");
  $err = -1;
  if (file_exists (HGREPOS . "/$repos[name]/.hg") == true) {
    passthru ("hg -R " . HGREPOS . "/$repos[name] verify >/dev/null", $err);
    logging (_LOG_DEBUG_, "passthru ('hg -R " . HGREPOS . "/$repos[name] verify', $err)");
  }
  if ($err != 0) {
    $msg .= logging (_LOG_ERROR_, "Directory '$repos[name]' is not a mercurial repository");
    return 2;
  }

  /* update status */
  $update_done = false;

  /* move directory */
  $name = (isset ($resources["name"])) ? $resources["name"] : "";
  if ((strlen ($name) != 0) && (strcmp ($repos["name"], $name) != 0)) {

    /* check name parameter */
    $path = real_name ($name);
    if (strlen ($path) == 0) {
      $msg .= logging (_LOG_ERROR_, "Path '$name' is empty or outside Mercurail global repository");
      return 2;
    }
    $name = $path;
    logging (_LOG_DEBUG_, "name: '$name'");

    /* check if new name exists */
    if (file_exists (HGREPOS . "/$name/.hg")) {
      $msg .= logging (_LOG_ERROR_, "Can't move directory '$repos[name]' to '$name' (repository already exist)");
      return 2;
    }

    /* check parent name */
    $dirname = dirname ($name);
    if (strcmp ($dirname, ".") == 0) $dirname = "";
    if (strlen ($dirname) != 0) {
      $ret = true;
      if (!file_exists (HGREPOS . "/$dirname")) {
        $ret = @mkdir (HGREPOS . "/$dirname", 0750, true);
        logging (_LOG_DEBUG_, "$ret = mkdir ('" . HGREPOS . "/$dirname', 0750, true)");
      } else if (!is_dir (HGREPOS . "/$dirname")) {
        $ret = false;
      }
      if (!$ret) {
        $msg .= logging (_LOG_ERROR_, "Can't create parent directory '$dirname'");
        return 2;
      }
    }

    $ret = false;
    if (!file_exists (HGREPOS . "/$name")) {
      /* move it straightforwardly */

      $ret = @rename (HGREPOS . "/$repos[name]", HGREPOS . "/$name");
      logging (_LOG_DEBUG_, "$ret = rename ('" . HGREPOS . "/$repos[name]', '" . HGREPOS . "/$name')");

    } else if (is_dir (HGREPOS . "/$name")) {
      /* move its contents */

      $dh = opendir (HGREPOS . "/$repos[name]");
      logging (_LOG_DEBUG_, "$dh = opendir ('" . HGREPOS . "/$repos[name]')");
      if (!$dh) {
        $msg .= logging (_LOG_ERROR_, "Can't open directory '$repos[name]'");
        return 2;
      }

      $ret = false;
      $list = array ();
      while (($f = readdir ($dh)) != false) {
        if ((strcmp ($f, ".") == 0) || (strcmp ($f, "..") == 0)) continue;

        $ret = @rename (HGREPOS . "/$repos[name]/$f", HGREPOS . "/$name/$f");
        logging (_LOG_DEBUG_, "$ret = rename ('" . HGREPOS . "/$repos[name]/$f', '" . HGREPOS . "/$name/$f')");

        if (!$ret)
          break;

        array_push ($list, $f);
      }

      if (!$ret) {
        /* on error move back all files */

        if (count ($list) != 0) {
          foreach ($list as $f) {
            $ret = @rename (HGREPOS . "/$name/$f", HGREPOS . "/$repos[name]/$f");
            logging (_LOG_DEBUG_, "$ret = rename ('" . HGREPOS . "/$name/$f', '" . HGREPOS . "/$repos[name]/$f')");
          }
        }

      } else {
        /* on success try to remove parent directory */

        $ret_rmdir = @rmdir (HGREPOS . "/$repos[name]");
        logging (_LOG_DEBUG_, "$ret_rmdir = rmdir ('" . HGREPOS . "/$repos[name]')");
        if (!$ret_rmdir) {
          $msg .= logging (_LOG_WARN_, "Can't remove old directory '$name'");
          $stat = 1;
        }
      }
    }

    /* global status of moving */
    if (!$ret) {
      $msg .= logging (_LOG_ERROR_, "Can't move '$repos[name]' to '$name'");
      return 2;
    }

    $msg .= logging (_LOG_DEBUG_, "Move '$repos[name]' to '$name'");
    $repos["name"] = $name;
    $update_done = true;
  }

  /* resource file */
  $old_resources = format_resource ($repos);
  logging (_LOG_DEBUG_, "Old resources for '$repos[name]': '$old_resources'");
  if (isset ($repos["rest"]))
      $resources["rest"] = $repos["rest"];
  $new_resources = format_resource (array_merge ($repos, $resources));
  logging (_LOG_DEBUG_, "New resources for '$repos[name]': '$new_resources'");
  if (strcmp ($old_resources, $new_resources) != 0) {
    switch (file_write (HGREPOS . "/$repos[name]/.hg/hgrc", $new_resources)) {
    case -1:
      $msg .= logging (_LOG_WARN_, "Can't write resource file for '$repos[name]'");
      $stat = 1;
      break;
    case 0:
      $msg .= logging (_LOG_DEBUG_, "Update resource file for '$repos[name]'");
      break;
    case 1:
      $msg .= logging (_LOG_WARN_, "Resource file for '$repos[name]' not completely updated");
      $stat = 1;
      break;
    }
    $update_done = true;
  }

  if (!$update_done) {
    $msg .= logging (_LOG_WARN_, "Nothing to update in '$repos[name]'");
    $stat = 1;
  }

  return $stat;
}

/** Main actions **/

$hgtree = build_hg_tree ();

/* action list */

Header("content-type: application/xml");

if (array_key_exists ("xmltree", $_GET)) {

  echo create_xml_tree ($hgtree);

} else if (array_key_exists ("xmlconfig", $_GET)) {

  echo create_xml_config ();

} else if (array_key_exists ("action", $_POST)) {

  /* initialisation */
  $name = "";
  $base = "";

  $msg = "";

  switch ($_POST["action"]) {

  case "config":

    $stat = update_configuration ($_POST, $msg);

    break;

  case "create":

    /* extract parameters */
    if (array_key_exists ("name", $_POST))
      $name = $_POST["name"];
    logging (_LOG_DEBUG_, "name = '$name'");

    $stat = create_repository ($hgtree, $name, $_POST, $msg);

    break;

  case "delete":

    /* extract parameters */
    if (array_key_exists ("base", $_POST))
      $base = $_POST["base"];
    logging (_LOG_DEBUG_, "base = '$base'");

    $stat = delete_repository ($hgtree, $base, $msg);

    break;

  case "edit":

    /* extract parameters */
    if (array_key_exists ("base", $_POST))
      $base = $_POST["base"];
    logging (_LOG_DEBUG_, "base = '$base'");

    $stat = edit_repository ($hgtree, $base, $_POST, $msg);

    break;

  default:

    $stat = 0;

  }

  $msg = preg_replace ("/\n*\$/", "", $msg);
  echo "<xml><message><text><![CDATA[$msg]]></text><stat>$stat</stat></message></xml>";
}

/* Local Variables:  */
/* mode: PHP         */
/* c-basic-offset: 2 */
/* End:              */
?>
